# Copyright (C) 2018-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import defaultdict
import copy
from typing import Any, Dict, List, Tuple

from django.http.request import HttpRequest
from django.http.response import HttpResponse
from django.shortcuts import render
from django.views.decorators.cache import cache_page
from django.views.decorators.clickjacking import xframe_options_exempt

from swh.scheduler.model import SchedulerMetrics
from swh.web.config import scheduler
from swh.web.utils import reverse
from swh.web.utils.archive import count_origins

# Current coverage list of the archive in a high level overview fashion,
# categorized as follow:
#   - listed origins: origins discovered using a swh lister
#   - legacy: origins where public hosting service has closed
#   - deposited: origins coming from swh-deposit
#
# TODO: Store that list in a database table somewhere (swh-scheduler, swh-storage ?)
#       and retrieve it dynamically
listed_origins: Dict[str, Any] = {
    "info": (
        "These software origins get continuously discovered and archived using "
        'the <a href="https://docs.softwareheritage.org/user/listers.html" target="_blank" '
        'rel="noopener noreferrer">listers</a> implemented by Software Heritage.'
    ),
    "origins": [
        {
            "type": "arch",
            "info_url": "https://archlinux.org/",
            "info": "source packages from the Arch Linux distribution",
            "search_pattern": {
                "default": "https://archlinux.org/packages/",
            },
        },
        {
            "type": "aur",
            "info_url": "https://aur.archlinux.org/",
            "info": "source packages from the Arch Linux User Repository",
            "search_pattern": {
                "default": "https://aur.archlinux.org/packages/",
            },
        },
        {
            "type": "bioconductor",
            "info_url": "https://bioconductor.org/",
            "info": "source packages from the Bioconductor project",
            "search_pattern": {
                "default": "https://bioconductor.org/",
            },
        },
        {
            "type": "bitbucket",
            "info_url": "https://bitbucket.org",
            "info": "public repositories from Bitbucket",
            "search_pattern": {
                "default": "https://bitbucket.org/",
            },
        },
        {
            "type": "bower",
            "info_url": "https://bower.io/",
            "info": (
                "git repositories of packages referenced by Bower, "
                "a package manager for the web"
            ),
            "search_pattern": {
                "default": "",
            },
        },
        {
            "type": "cgit",
            "info_url": "https://git.zx2c4.com/cgit/about",
            "info": "public repositories from cgit instances",
            "search_pattern": {
                "default": "cgit",
            },
        },
        {
            "type": "conda",
            "info_url": "https://conda.io/",
            "info": (
                "source packages from the Conda open source package management "
                "system and environment management"
            ),
            "search_pattern": {
                "default": "https://anaconda.org/",
            },
        },
        {
            "type": "cpan",
            "info_url": "https://www.cpan.org/",
            "info": "source packages from The Comprehensive Perl Archive Network",
            "search_pattern": {
                "default": "https://metacpan.org/",
            },
        },
        {
            "type": "cran",
            "info_url": "https://cran.r-project.org",
            "info": "source packages from The Comprehensive R Archive Network",
            "search_pattern": {
                "default": "https://cran.r-project.org/",
            },
        },
        {
            "type": "crates",
            "info_url": "https://crates.io/",
            "info": "source packages from The Rust community's crate registry",
            "search_pattern": {
                "default": "https://crates.io/crates/",
            },
        },
        {
            "type": "dlang",
            "info_url": "https://dlang.org/",
            "info": "public repositories of packages for the D programming language",
            "search_pattern": {
                "default": "",
            },
        },
        {
            "type": "debian",
            "info_url": "https://www.debian.org",
            "info": "source packages from Debian and Debian-based distributions",
            "search_pattern": {
                "default": "deb://",
            },
        },
        {
            "type": "forgejo",
            "info_url": "https://forgejo.org/",
            "info": "public repositories from Forgejo instances",
            "search_pattern": {
                "default": "forgejo",
            },
        },
        {
            "type": "gitea",
            "info_url": "https://about.gitea.com/",
            "info": "public repositories from Gitea instances",
            "search_pattern": {
                "default": "gitea",
            },
        },
        {
            "type": "github",
            "info_url": "https://github.com",
            "info": "public repositories from GitHub",
            "search_pattern": {
                "default": "https://github.com/",
            },
        },
        {
            "type": "gitiles",
            "info_url": "https://github.com/google/gitiles",
            "info": "public repositories from Gitiles instances",
            "search_pattern": {
                "default": "gitiles",
            },
        },
        {
            "type": "gitlab",
            "info_url": "https://gitlab.com",
            "info": "public repositories from multiple GitLab instances",
            "search_pattern": {
                "default": "gitlab",
            },
        },
        {
            "type": "gitweb",
            "info_url": "https://git-scm.com/book/en/v2/Git-on-the-Server-GitWeb",
            "info": "public repositories from GitWeb instances",
            "search_pattern": {
                "default": "gitweb",
            },
        },
        {
            "type": "gogs",
            "info_url": "https://gogs.io/",
            "info": "public repositories from multiple Gogs instances",
            "search_pattern": {
                "default": "gogs",
            },
        },
        {
            "type": "golang",
            "info_url": "https://go.dev/",
            "info": "source packages from the official Go package manager",
            "search_pattern": {
                "default": "https://pkg.go.dev/",
            },
        },
        {
            "type": "guix",
            "info_url": "https://guix.gnu.org",
            "info": "source code tarballs used to build the Guix package collection",
            "main_instance": "guix.gnu.org",
            "search_pattern": {
                "default": "",
            },
        },
        {
            "type": "GNU",
            "info_url": "https://www.gnu.org",
            "info": "releases from the GNU project (as of August 2015)",
            "search_pattern": {
                "default": "gnu",
            },
        },
        {
            "type": "hackage",
            "info_url": "https://hackage.haskell.org",
            "info": "source packages from The Haskell Package Repository",
            "search_pattern": {
                "default": "https://hackage.haskell.org/package/",
            },
        },
        {
            "type": "heptapod",
            "info_url": "https://heptapod.net/",
            "info": "public repositories from multiple Heptapod instances",
            "search_pattern": {
                "default": "heptapod",
            },
        },
        {
            "type": "launchpad",
            "info_url": "https://launchpad.net",
            "logo": "img/logos/launchpad.png",
            "info": "public repositories from Launchpad",
            "search_pattern": {
                "default": "launchpad.net/",
            },
        },
        {
            "type": "maven",
            "info_url": "https://maven.apache.org/",
            "info": "java source packages from maven repositories",
            "search_pattern": {
                "default": "maven",
                "cvs": "",
                "git": "",
                "hg": "",
                "svn": "",
            },
        },
        {
            "type": "nixos",
            "info_url": "https://nixos.org",
            "info": "source code tarballs and patches used to build the Nix package collection",
            "main_instance": "nixpkgs-swh.nixos.org",
            "search_pattern": {
                "default": "",
                "content": "https://cache.nixos.org",
                "tarball-directory": "https://cache.nixos.org",
            },
        },
        {
            "type": "npm",
            "info_url": "https://www.npmjs.com",
            "info": "public packages from the package registry for javascript",
            "search_pattern": {
                "default": "https://www.npmjs.com",
            },
        },
        {
            "type": "opam",
            "info_url": "https://opam.ocaml.org/",
            "info": "public packages from the source-based package manager for OCaml",
            "search_pattern": {
                "default": "opam+https://",
            },
        },
        {
            "type": "Packagist",
            "info_url": "https://packagist.org/",
            "info": "source code repositories referenced by The PHP Package Repository",
            "search_pattern": {
                "default": "",
            },
        },
        {
            "type": "pagure",
            "info_url": "https://pagure.io/pagure",
            "info": "public repositories from multiple Pagure instances",
            "search_pattern": {
                "default": "",
            },
        },
        {
            "type": "phabricator",
            "info_url": "https://www.phacility.com/phabricator",
            "info": "public repositories from multiple Phabricator instances",
            "search_pattern": {
                "default": "phabricator",
            },
        },
        {
            "type": "pubdev",
            "info_url": "https://pub.dev",
            "info": "source packages from the official repository for Dart and Flutter apps",
            "search_pattern": {
                "default": "https://pub.dev",
            },
        },
        {
            "type": "puppet",
            "info_url": "https://forge.puppet.com/",
            "info": "source packages from the Puppet Forge",
            "search_pattern": {
                "default": "https://forge.puppet.com/modules/",
            },
        },
        {
            "type": "pypi",
            "info_url": "https://pypi.org",
            "info": "source packages from the Python Package Index",
            "search_pattern": {
                "default": "https://pypi.org",
            },
        },
        {
            "type": "rpm",
            "info_url": "https://www.redhat.com",
            "info": "source packages from Red Hat based distributions",
            "search_pattern": {
                "default": "rpm",
            },
        },
        {
            "type": "rubygems",
            "info_url": "https://rubygems.org",
            "info": "source packages from the Ruby community's gem hosting service",
            "search_pattern": {
                "default": "https://rubygems.org/gems/",
            },
        },
        {
            "type": "sourceforge",
            "info_url": "https://sourceforge.net",
            "info": "public repositories from SourceForge",
            "search_pattern": {
                "default": "code.sf.net",
                "bzr": "bzr.sourceforge.net",
                "cvs": "cvs.sourceforge.net",
            },
        },
        {
            "type": "stagit",
            "info_url": "https://codemadness.org/stagit.html",
            "info": "public repositories from Stagit instances",
            "search_pattern": {
                "default": "stagit",
            },
        },
    ],
}

legacy_origins: Dict[str, Any] = {
    "info": (
        "Discontinued hosting services. Those origins have been archived "
        "by Software Heritage."
    ),
    "origins": [
        {
            "type": "bitbucket-hg",
            "info_url": "https://bitbucket.org",
            "info": "public mercurial repositories from Bitbucket",
            "search_pattern": "https://bitbucket.org/",
            "visit_types": ["hg"],
        },
        {
            "type": "gitorious",
            "info_url": "https://en.wikipedia.org/wiki/Gitorious",
            "info": (
                "public repositories from the former Gitorious code hosting service"
            ),
            "visit_types": ["git"],
            "search_pattern": "https://gitorious.org",
        },
        {
            "type": "googlecode",
            "info_url": "https://code.google.com/archive",
            "info": (
                "public repositories from the former Google Code project "
                "hosting service"
            ),
            "visit_types": ["git", "hg", "svn"],
            "search_pattern": "googlecode.com",
        },
        {
            "type": "osdn",
            "info_url": "https://en.wikipedia.org/wiki/OSDN",
            "info": (
                "public repositories from the former Open Source Development Network "
                "hosting service"
            ),
            "visit_types": ["cvs", "git", "hg", "svn"],
            "search_pattern": "osdn.net",
        },
    ],
}

deposited_origins: Dict[str, Any] = {
    "info": (
        "These origins are directly pushed into the archive by trusted partners "
        "using the "
        '<a href="https://docs.softwareheritage.org/devel/architecture/overview.html#deposit" '
        'target="_blank" rel="noopener noreferrer">deposit</a> service of Software Heritage.'
    ),
    "origins": [
        {
            "type": "elife",
            "info_url": "https://elifesciences.org",
            "info": (
                "research software source code associated to the articles "
                "eLife publishes"
            ),
            "search_pattern": "elife.stencila.io",
            "visit_types": ["deposit"],
        },
        {
            "type": "hal",
            "info_url": "https://hal.archives-ouvertes.fr",
            "info": "scientific software source code deposited in the open archive HAL",
            "visit_types": ["deposit"],
            "search_pattern": "hal.archives-ouvertes.fr",
        },
        {
            "type": "ipol",
            "info_url": "https://www.ipol.im",
            "info": "software artifacts associated to the articles IPOL publishes",
            "visit_types": ["deposit"],
            "search_pattern": "doi.org/10.5201/ipol",
        },
        {
            "type": "zenodo",
            "info_url": "https://zenodo.org/",
            "info": "software source code deposited in the Open Science platform Zenodo",
            "visit_types": ["deposit"],
            "search_pattern": "doi.org/10.5281/zenodo",
        },
    ],
}


def get_listers_metrics() -> Dict[str, List[Tuple[str, SchedulerMetrics]]]:
    """Returns scheduler metrics in the following mapping:
    Dict[lister_name, List[Tuple[instance_name, SchedulerMetrics]]]
    as a lister instance has one SchedulerMetrics object per visit type.
    """
    listers_metrics = defaultdict(list)
    listers = scheduler().get_listers()
    scheduler_metrics = scheduler().get_metrics()
    for lister in listers:
        for metrics in filter(lambda m: m.lister_id == lister.id, scheduler_metrics):
            listers_metrics[lister.name].append((lister.instance_name, metrics))

    return listers_metrics


def _search_url(query: str, visit_type: str) -> str:
    return reverse(
        "browse-search",
        query_params={
            "q": query,
            "visit_type": visit_type,
            "with_visit": "true",
            "with_content": "true",
        },
    )


@xframe_options_exempt
@cache_page(24 * 60 * 60)  # one day
def swh_coverage(request: HttpRequest) -> HttpResponse:
    listers_metrics = get_listers_metrics()

    for origins in listed_origins["origins"]:
        origins["count"] = "0"
        origins["instances"] = {}
        origins_type = origins["type"]

        # special processing for nixos/guix origins
        if origins_type in ("nixos", "guix"):
            main_instance = origins["main_instance"]
            visit_data = listers_metrics.get("nixguix", [])
            visit_type_counts: Dict[str, Dict[str, Dict[str, str]]] = defaultdict(dict)
            # visit types from new nixguix lister
            count = 0
            for instance, metrics in visit_data:
                if instance != main_instance:
                    continue
                instance_type_count = (
                    metrics.origins_enabled - metrics.origins_never_visited
                )
                if instance_type_count > 0:
                    visit_type_counts[instance][metrics.visit_type] = {
                        "count": f"{instance_type_count:,}"
                    }

                count += instance_type_count

            origins["count"] = f"{count:,}"
            origins["instances"] = dict(visit_type_counts)

        if origins_type not in listers_metrics:
            continue

        count_total = sum(
            [metrics.origins_enabled for _, metrics in listers_metrics[origins_type]]
        )
        count_never_visited = sum(
            [
                metrics.origins_never_visited
                for _, metrics in listers_metrics[origins_type]
            ]
        )
        count_visited = count_total - count_never_visited

        origins["count"] = f"{count_visited:,}"
        origins["instances"] = defaultdict(dict)
        for instance, metrics in listers_metrics[origins_type]:
            instance_count = metrics.origins_enabled - metrics.origins_never_visited
            # no archived origins for that visit type, skip it
            if instance_count == 0:
                continue

            origins["instances"][instance].update(
                {metrics.visit_type: {"count": f"{instance_count:,}"}}
            )
            origins["visit_types"] = list(
                set(origins["instances"][instance].keys())
                | set(origins.get("visit_types", []))
            )

        # defaultdict cannot be iterated in django template
        origins["instances"] = dict(sorted(origins["instances"].items()))

    for origins in listed_origins["origins"]:
        instances = origins["instances"]
        nb_instances = len(instances)
        for instance_name, visit_types in instances.items():
            for visit_type in visit_types:
                search_url = ""
                if visit_type in origins["search_pattern"]:
                    search_pattern = origins["search_pattern"][visit_type]
                elif nb_instances > 1:
                    search_pattern = instance_name
                else:
                    search_pattern = origins["search_pattern"]["default"]
                if search_pattern:
                    search_url = _search_url(search_pattern, visit_type)
                visit_types[visit_type]["search_url"] = search_url

    # use a light copy to avoid side effects when running tests
    listed_origins_filtered = copy.copy(listed_origins)
    # filter out origin types without archived origins
    listed_origins_filtered["origins"] = list(
        filter(lambda o: o["count"] != "0", listed_origins_filtered["origins"])
    )

    for origins in legacy_origins["origins"]:
        origins["instances"] = {origins["type"]: {}}
        total_count = 0
        for visit_type in origins["visit_types"]:
            count = count_origins(
                origins["search_pattern"], with_visit=True, visit_types=[visit_type]
            )
            total_count += count
            origins["instances"][origins["type"]][visit_type] = {
                "count": f"{count:,}",
                "search_url": _search_url(origins["search_pattern"], visit_type),
            }
        origins["count"] = f"{total_count:,}"

    for origins in deposited_origins["origins"]:
        origins_count = count_origins(
            origins["search_pattern"], with_visit=True, visit_types=["deposit"]
        )
        origins["count"] = f"{origins_count:,}"
        origins["search_urls"] = {
            "deposit": _search_url(origins["search_pattern"], "deposit")
        }

    focus = []
    focus_param = request.GET.get("focus")
    if focus_param:
        focus = focus_param.split(",")

    return render(
        request,
        "archive-coverage.html",
        {
            "origins": {
                "Regular crawling": listed_origins_filtered,
                "Discontinued hosting": legacy_origins,
                "On demand archival": deposited_origins,
            },
            "focus": focus,
        },
    )



# Copyright (C) 2018-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from swh.web.utils.exc import handle_view_exception


class ThrottlingHeadersMiddleware(object):
    """
    Django middleware for inserting rate limiting related
    headers in HTTP response.
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        resp = self.get_response(request)
        if "RateLimit-Limit" in request.META:
            resp["X-RateLimit-Limit"] = request.META["RateLimit-Limit"]
        if "RateLimit-Remaining" in request.META:
            resp["X-RateLimit-Remaining"] = request.META["RateLimit-Remaining"]
        if "RateLimit-Reset" in request.META:
            resp["X-RateLimit-Reset"] = request.META["RateLimit-Reset"]
        return resp


class ExceptionMiddleware(object):
    """
    Django middleware for handling uncaught exception raised when
    processing a view.
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    def process_exception(self, request, exception):
        return handle_view_exception(request, exception)



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


import rfc3987

from rest_framework import serializers

from swh.model.exceptions import ValidationError
from swh.model.swhids import QualifiedSWHID


class SoftLimitsIntegerField(serializers.IntegerField):
    """An IntegerField that soft-enforce min/max values."""

    def to_internal_value(self, data) -> int:
        """Before returning `data` limit its value to min_value & max_value.

        Args:
            data: a string representation of a value

        Returns:
            An integer between min_value & max_value
        """
        data = super().to_internal_value(data)
        if self.min_value:
            data = max(data, self.min_value)
        if self.max_value:
            data = min(data, self.max_value)
        return data


class SWHIDField(serializers.CharField):
    """A SWHID field."""

    def to_internal_value(self, data: str) -> str:
        """Validate a SWHID.

        Args:
            value: a string

        Raises:
            serializers.ValidationError: invalid SWHID

        Returns:
            a string
        """
        data = super().to_internal_value(data)
        try:
            QualifiedSWHID.from_string(data)
        except ValidationError as e:
            raise serializers.ValidationError(
                f"Invalid target SWHID {data}: {e.message}."
            )
        return data


class IRIField(serializers.CharField):
    """An Internationalized Resource Identifier field.

    Requires a valid IRI value and limits its length to 4096, can be used to handle
    origin URLS.
    """

    def __init__(self, **kwargs):
        """Defaults max_length to 4096."""
        if "max_length" not in kwargs:
            kwargs["max_length"] = 4096
        super().__init__(**kwargs)

    def to_internal_value(self, data: str) -> str:
        """Validate the IRI.

        Args:
            data: a string

        Raises:
            serializers.ValidationError: invalid IRI

        Returns:
            an URI
        """
        # white spaces must be escaped in an IRI
        data = super().to_internal_value(data).replace(" ", "%20")
        try:
            rfc3987.parse(data, rule="IRI")
        except ValueError as e:
            raise serializers.ValidationError(str(e))
        return data



# Copyright (C) 2022-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import attr
from hypothesis import given, settings
from hypothesis.strategies import sets
import pytest

from swh.model.hypothesis_strategies import raw_extrinsic_metadata
from swh.model.model import Origin
from swh.web.api.tests.views.utils import scroll_results
from swh.web.tests.helpers import check_api_get_responses, check_http_get_response
from swh.web.utils import reverse


@given(raw_extrinsic_metadata())
def test_api_raw_extrinsic_metadata(api_client, subtest, metadata):
    # ensure archive_data fixture will be reset between each hypothesis
    # example test run
    @subtest
    def test_inner(archive_data):
        archive_data.metadata_authority_add([metadata.authority])
        archive_data.metadata_fetcher_add([metadata.fetcher])
        archive_data.raw_extrinsic_metadata_add([metadata])

        authority = metadata.authority
        url = reverse(
            "api-1-raw-extrinsic-metadata-swhid",
            url_args={"target": str(metadata.target)},
            query_params={"authority": f"{authority.type.value} {authority.url}"},
        )
        rv = check_api_get_responses(api_client, url, status_code=200)

        assert len(rv.data) == 1

        expected_result = metadata.to_dict()
        del expected_result["id"]
        del expected_result["metadata"]
        metadata_url = rv.data[0]["metadata_url"]
        expected_result["metadata_url"] = metadata_url
        expected_result["discovery_date"] = expected_result[
            "discovery_date"
        ].isoformat()
        if expected_result["target"].startswith(("swh:1:ori:", "swh:1:emd:")):
            # non-core SWHID are hidden from the API
            del expected_result["target"]
        assert rv.data == [expected_result]

        rv = check_http_get_response(api_client, metadata_url, status_code=200)
        assert rv["Content-Type"] == "application/octet-stream"
        assert (
            rv["Content-Disposition"]
            == f'attachment; filename="{metadata.target}_metadata"'
        )
        assert rv.content == metadata.metadata


@settings(max_examples=1)
@given(raw_extrinsic_metadata())
def test_api_raw_extrinsic_metadata_origin_filename(api_client, subtest, metadata):
    # ensure archive_data fixture will be reset between each hypothesis
    # example test run
    @subtest
    def test_inner(archive_data):
        nonlocal metadata
        origin = Origin(url="http://example.com/repo.git")
        metadata = attr.evolve(metadata, target=origin.swhid())
        metadata = attr.evolve(metadata, id=metadata.compute_hash())
        archive_data.origin_add([origin])
        archive_data.metadata_authority_add([metadata.authority])
        archive_data.metadata_fetcher_add([metadata.fetcher])
        archive_data.raw_extrinsic_metadata_add([metadata])

        authority = metadata.authority
        url = reverse(
            "api-1-raw-extrinsic-metadata-swhid",
            url_args={"target": str(metadata.target)},
            query_params={"authority": f"{authority.type.value} {authority.url}"},
        )
        rv = check_api_get_responses(api_client, url, status_code=200)

        assert len(rv.data) == 1
        metadata_url = rv.data[0]["metadata_url"]
        rv = check_http_get_response(api_client, metadata_url, status_code=200)
        assert rv["Content-Type"] == "application/octet-stream"
        assert (
            rv["Content-Disposition"]
            == 'attachment; filename="http_example_com_repo_git_metadata"'
        )
        assert rv.content == metadata.metadata


@pytest.mark.parametrize("limit", [1, 2, 10, 100])
@given(sets(raw_extrinsic_metadata(), min_size=1))
def test_api_raw_extrinsic_metadata_scroll(api_client, subtest, limit, meta):
    # ensure archive_data fixture will be reset between each hypothesis
    # example test run
    @subtest
    def test_inner(archive_data):
        # Make all metadata objects use the same authority and target
        metadata0 = next(iter(meta))
        metadata = {
            attr.evolve(m, authority=metadata0.authority, target=metadata0.target)
            for m in meta
        }
        # Metadata ids must also be updated as they depend on authority and target
        metadata = {attr.evolve(m, id=m.compute_hash()) for m in metadata}
        authority = metadata0.authority

        archive_data.metadata_authority_add([authority])
        archive_data.metadata_fetcher_add(list({m.fetcher for m in metadata}))
        archive_data.raw_extrinsic_metadata_add(metadata)
        url = reverse(
            "api-1-raw-extrinsic-metadata-swhid",
            url_args={"target": str(metadata0.target)},
            query_params={
                "authority": f"{authority.type.value} {authority.url}",
                "limit": limit,
            },
        )

        results = scroll_results(api_client, url)

        expected_results = [m.to_dict() for m in metadata]

        for expected_result in expected_results:
            del expected_result["id"]
            del expected_result["metadata"]
            expected_result["discovery_date"] = expected_result[
                "discovery_date"
            ].isoformat()
            if expected_result["target"].startswith(("swh:1:ori:", "swh:1:emd:")):
                # non-core SWHID are hidden from the API
                del expected_result["target"]

        assert len(results) == len(expected_results)

        for result in results:
            del result["metadata_url"]
            assert result in expected_results, str(expected_results)


_swhid = "swh:1:dir:a2faa28028657859c16ff506924212b33f0e1307"


@pytest.mark.parametrize(
    "status_code,url_args,query_params",
    [
        pytest.param(
            200,
            {"target": _swhid},
            {"authority": "forge http://example.org"},
            id="minimal working",
        ),
        pytest.param(
            200,
            {"target": _swhid},
            {
                "authority": "forge http://example.org",
                "after": "2021-06-18T09:31:09",
                "limit": 100,
            },
            id="maximal working",
        ),
        pytest.param(
            400,
            {"target": _swhid},
            {"authority": "foo http://example.org"},
            id="invalid authority type",
        ),
        pytest.param(
            400,
            {"target": _swhid},
            {
                "authority": "forge http://example.org",
                "after": "yesterday",
            },
            id="invalid 'after' format",
        ),
        pytest.param(
            400,
            {"target": _swhid},
            {
                "authority": "forge http://example.org",
                "limit": "abc",
            },
            id="invalid 'limit'",
        ),
    ],
)
def test_api_raw_extrinsic_metadata_check_params(
    api_client, archive_data, status_code, url_args, query_params
):
    url = reverse(
        "api-1-raw-extrinsic-metadata-swhid",
        url_args=url_args,
        query_params=query_params,
    )
    check_api_get_responses(api_client, url, status_code=status_code)


@settings(max_examples=1)
@given(raw_extrinsic_metadata())
def test_api_raw_extrinsic_metadata_list_authorities(
    api_client, archive_data, metadata
):
    archive_data.metadata_authority_add([metadata.authority])
    archive_data.metadata_fetcher_add([metadata.fetcher])
    archive_data.raw_extrinsic_metadata_add([metadata])

    authority = metadata.authority
    url = reverse(
        "api-1-raw-extrinsic-metadata-swhid-authorities",
        url_args={"target": str(metadata.target)},
    )
    rv = check_api_get_responses(api_client, url, status_code=200)

    expected_results = [
        {
            "type": authority.type.value,
            "url": authority.url,
            "metadata_list_url": "http://testserver"
            + reverse(
                "api-1-raw-extrinsic-metadata-swhid",
                url_args={"target": str(metadata.target)},
                query_params={"authority": f"{authority.type.value} {authority.url}"},
            ),
        }
    ]

    assert rv.data == expected_results


def test_api_raw_extrinsic_metadata_origin_redirect(api_client, archive_data):
    origin = Origin(url="http://example.com/repo.git")
    archive_data.origin_add([origin])

    url = reverse(
        "api-1-raw-extrinsic-metadata-origin-authorities",
        url_args={"origin_url": origin.url},
    )
    rv = check_http_get_response(api_client, url, status_code=302)

    redirect_url = reverse(
        "api-1-raw-extrinsic-metadata-swhid-authorities",
        url_args={"target": str(origin.swhid())},
    )

    assert rv["location"] == redirect_url



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest
from pytest_django.asserts import assertNumQueries

from django.forms.models import model_to_dict

from swh.web.alter.forms import (
    AlterationAccessForm,
    AlterationAdminForm,
    AlterationForm,
    AlterationSearchForm,
    EmailVerificationForm,
    EventAdminForm,
    MessageAdminForm,
    MessageForm,
    OriginAdminForm,
    OriginSelectForm,
)
from swh.web.alter.models import (
    AlterationStatus,
    BlockList,
    EventCategory,
    EventRecipient,
    Token,
)


def test_origin_select_form_ko(origin):
    data = {"urls": ["http://an.unknown.origin", origin["url"]]}
    form = OriginSelectForm(data)
    assert not form.is_valid()
    errors = form["urls"].errors
    assert len(errors) == 1
    assert "http://an.unknown.origin" in errors[0]


def test_origin_select_form_ok(origin):
    data = {"urls": [origin["url"]]}
    form = OriginSelectForm(data)
    assert form.is_valid()


def test_alteration_access_form_ok(alteration, mailoutbox, rf):
    assert alteration.tokens.count() == 0
    data = {"email": alteration.email}
    form = AlterationAccessForm(data, alteration=alteration, request=rf.get("/"))
    assert form.is_valid()
    assert len(mailoutbox) == 1
    message = mailoutbox[0]
    assert alteration.tokens.count() == 1
    token = alteration.tokens.first()
    assert token.get_absolute_url() in message.body


def test_alteration_access_form_ko(alteration, mailoutbox, rf):
    assert alteration.tokens.count() == 0
    data = {"email": "wrong@email.address"}
    form = AlterationAccessForm(data, alteration=alteration, request=rf.get("/"))
    assert form.is_valid()  # the form is still valid even if the email doesn't match
    assert len(mailoutbox) == 0
    assert alteration.tokens.count() == 0


def test_alteration_search_form(alteration):
    form = AlterationSearchForm({})
    assert form.is_valid()
    page = form.search()
    with assertNumQueries(1):  # related origins are fetched in a single query
        first_result = page.object_list.first()
        assert first_result == alteration
        assert first_result.origins == alteration.origins


@pytest.mark.parametrize(
    "query,status,expected_count",
    [
        ("ZZZ", AlterationStatus.VALIDATING, 0),
        ("requester", AlterationStatus.VALIDATING, 1),
        ("requester", AlterationStatus.CLOSED, 0),
        (None, AlterationStatus.VALIDATING, 1),
        (None, AlterationStatus.CLOSED, 0),
        ("requester", None, 1),
    ],
)
def test_alteration_search_form_filters(alteration, query, status, expected_count):
    data = {"query": query, "status": status}
    form = AlterationSearchForm(data)
    assert form.is_valid()
    page = form.search()
    assert page.object_list.count() == expected_count


def test_message_form_ok(alteration):
    previous_count = alteration.events.count()
    data = {"content": "My message", "internal": True}  # internal will be ignored
    form = MessageForm(data, alteration=alteration)
    assert form.is_valid()
    event = form.save()
    assert alteration.events.count() == previous_count + 1
    assert alteration.events.first() == event
    assert event.author == "Requester"
    assert event.category == EventCategory.MESSAGE
    assert event.recipient == EventRecipient.SUPPORT
    assert not event.internal
    assert event.content == "My message"


def test_message_form_ko(alteration):
    data = {"content": ""}
    form = MessageForm(data, alteration=alteration)
    assert not form.is_valid()


def test_message_admin_form_ok(alteration):
    previous_count = alteration.events.count()
    data = {
        "content": "My admin message",
        "internal": True,
        "recipient": EventRecipient.MANAGER,
    }
    form = MessageAdminForm(data, alteration=alteration, author="Admin")
    assert form.is_valid()
    event = form.save()
    assert alteration.events.count() == previous_count + 1
    assert alteration.events.first() == event
    assert event.author == "Admin"
    assert event.category == EventCategory.MESSAGE
    assert event.recipient == EventRecipient.MANAGER
    assert event.internal
    assert event.content == "My admin message"


def test_message_admin_form_internal_to_requestor(alteration):
    data = {
        "content": "Internal message to requester",
        "internal": True,
        "recipient": EventRecipient.REQUESTER,
    }
    form = MessageAdminForm(data, alteration=alteration, author="Admin")
    assert not form.is_valid()
    assert form["internal"].errors


def test_origin_admin_form(rf, alteration, staff_user):
    previous_count = alteration.events.count()
    request = rf.get("/admin/alteration")
    request.user = staff_user
    origin = alteration.origins.first()
    data = model_to_dict(origin)
    data["code_license"] = "to confuse"
    form = OriginAdminForm(data, instance=origin, request=request)
    assert form.is_valid()
    origin = form.save()
    assert origin.code_license == "to confuse"
    assert alteration.events.count() == previous_count + 1
    event = alteration.events.first()
    assert event.author == staff_user.username
    assert event.category == EventCategory.LOG
    assert not event.internal
    assert str(origin) in event.content
    assert "- license found in code: to confuse" in event.content


def test_origin_admin_form_no_change(alteration):
    previous_count = alteration.events.count()
    origin = alteration.origins.first()
    data = model_to_dict(origin)
    form = OriginAdminForm(data, instance=origin)
    assert form.is_valid()
    origin = form.save()
    assert alteration.events.count() == previous_count


def test_alteration_form(alteration):
    previous_count = alteration.events.count()
    data = model_to_dict(alteration)
    data["email"] = "new@email.address"  # will be ignored
    data["reasons"] = "new reasons"
    form = AlterationForm(data, author="Requester", instance=alteration)
    assert form.is_valid()
    alteration = form.save()
    assert alteration.email != "new@email.address"
    assert alteration.events.count() == previous_count + 1
    event = alteration.events.first()
    assert event.author == "Requester"
    assert event.category == EventCategory.LOG
    assert not event.internal
    assert "reasons: new reasons" in event.content
    assert "email" not in event.content


def test_alteration_form_no_change(alteration):
    previous_count = alteration.events.count()
    data = model_to_dict(alteration)
    form = AlterationForm(data, author="Requester", instance=alteration)
    assert form.is_valid()
    alteration = form.save()
    assert alteration.events.count() == previous_count


def test_admin_alteration_form(alteration):
    previous_count = alteration.events.count()
    data = model_to_dict(alteration)
    data["email"] = "new@email.address"
    form = AlterationAdminForm(data, author="Admin", instance=alteration)
    assert form.is_valid()
    alteration = form.save()
    assert alteration.email == "new@email.address"
    assert alteration.events.count() == previous_count + 1
    event = alteration.events.first()
    assert event.author == "Admin"
    assert event.category == EventCategory.LOG
    assert not event.internal
    assert "email" in event.content


@pytest.mark.django_db
def test_email_verification_form_ok(mailoutbox, rf):
    email = "test@email.address"
    assert not Token.objects.filter(email=email).exists()
    data = {"email": email}
    form = EmailVerificationForm(data, request=rf.get("/"))
    assert form.is_valid()

    assert Token.objects.filter(email=email).exists()
    assert len(mailoutbox) == 1
    message = mailoutbox[0]
    assert message.subject == "Please confirm your email address"


@pytest.mark.django_db
def test_email_verification_form_blocked(mailoutbox, rf):
    email = "test@email.address"
    BlockList.objects.create(email_or_domain=email)
    data = {"email": email}
    form = EmailVerificationForm(data, request=rf.get("/"))
    assert not form.is_valid()

    errors = form["email"].errors
    assert len(errors) == 1
    assert "has been blocked" in errors[0]

    assert len(mailoutbox) == 0
    assert not Token.objects.filter(email=email).exists()


def test_event_admin_form(alteration):
    event = alteration.events.first()
    data = model_to_dict(event)
    data["content"] = "new content"
    form = EventAdminForm(data, instance=event)
    assert form.is_valid()
    event = form.save()
    assert event.content == "new content"



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("swh_web_auth", "0003_delete_oidcuser"),
    ]

    operations = [
        migrations.CreateModel(
            name="UserMailmap",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("user_id", models.CharField(max_length=50, null=True)),
                ("from_email", models.TextField(unique=True, null=False)),
                ("from_email_verified", models.BooleanField(default=False)),
                (
                    "from_email_verification_request_date",
                    models.DateTimeField(null=True),
                ),
                ("display_name", models.TextField(null=False)),
                ("display_name_activated", models.BooleanField(default=False)),
                ("to_email", models.TextField(null=True)),
                ("to_email_verified", models.BooleanField(default=False)),
                ("to_email_verification_request_date", models.DateTimeField(null=True)),
                ("mailmap_last_processing_date", models.DateTimeField(null=True)),
                ("last_update_date", models.DateTimeField(auto_now=True)),
            ],
            options={
                "db_table": "user_mailmap",
            },
        ),
    ]



# Copyright (C) 2017-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random

from hypothesis import given
import pytest

from django.utils.html import escape

from swh.model.from_disk import DentryPerms
from swh.model.hashutil import hash_to_bytes, hash_to_hex
from swh.model.model import (
    Directory,
    DirectoryEntry,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Release,
    ReleaseTargetType,
    Revision,
    RevisionType,
    Snapshot,
    SnapshotBranch,
    SnapshotTargetType,
    TimestampWithTimezone,
)
from swh.model.swhids import ObjectType
from swh.storage.utils import now
from swh.web.browse.snapshot_context import process_snapshot_branches
from swh.web.tests.django_asserts import assert_contains, assert_not_contains
from swh.web.tests.helpers import check_html_get_response, check_http_get_response
from swh.web.tests.strategies import new_person, new_swh_date
from swh.web.utils import gen_path_info, reverse
from swh.web.utils.identifiers import gen_swhid


def test_root_directory_view(client, archive_data, directory):
    _directory_view_checks(client, directory, archive_data.directory_ls(directory))


def test_sub_directory_view(client, archive_data, directory_with_subdirs):
    dir_content = archive_data.directory_ls(directory_with_subdirs)
    subdir = random.choice([e for e in dir_content if e["type"] == "dir"])
    subdir_content = archive_data.directory_ls(subdir["target"])
    _directory_view_checks(
        client, directory_with_subdirs, subdir_content, subdir["name"]
    )


@given(new_person(), new_swh_date())
def test_sub_directory_view_origin_context(
    client, archive_data, empty_directory, person, date
):
    origin_url = "test_sub_directory_view_origin_context"
    subdir = Directory(
        entries=(
            DirectoryEntry(
                name=b"foo",
                type="dir",
                target=hash_to_bytes(empty_directory),
                perms=DentryPerms.directory,
            ),
            DirectoryEntry(
                name=b"bar",
                type="dir",
                target=hash_to_bytes(empty_directory),
                perms=DentryPerms.directory,
            ),
        )
    )

    parentdir = Directory(
        entries=(
            DirectoryEntry(
                name=b"baz",
                type="dir",
                target=subdir.id,
                perms=DentryPerms.directory,
            ),
        )
    )
    archive_data.directory_add([subdir, parentdir])

    revision = Revision(
        directory=parentdir.id,
        author=person,
        committer=person,
        message=b"commit message",
        date=TimestampWithTimezone.from_datetime(date),
        committer_date=TimestampWithTimezone.from_datetime(date),
        synthetic=False,
        type=RevisionType.GIT,
    )
    archive_data.revision_add([revision])

    snapshot = Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target="refs/head/master".encode(),
                target_type=SnapshotTargetType.ALIAS,
            ),
            b"refs/head/master": SnapshotBranch(
                target=revision.id,
                target_type=SnapshotTargetType.REVISION,
            ),
        }
    )
    archive_data.snapshot_add([snapshot])

    archive_data.origin_add([Origin(url=origin_url)])
    date = now()
    visit = OriginVisit(origin=origin_url, date=date, type="git")
    visit = archive_data.origin_visit_add([visit])[0]
    visit_status = OriginVisitStatus(
        origin=origin_url,
        visit=visit.visit,
        date=date,
        status="full",
        snapshot=snapshot.id,
    )
    archive_data.origin_visit_status_add([visit_status])

    dir_content = archive_data.directory_ls(hash_to_hex(parentdir.id))
    subdir = dir_content[0]
    subdir_content = archive_data.directory_ls(subdir["target"])
    _directory_view_checks(
        client,
        hash_to_hex(parentdir.id),
        subdir_content,
        subdir["name"],
        origin_url,
        hash_to_hex(snapshot.id),
        hash_to_hex(revision.id),
    )


def test_directory_request_errors(client, invalid_sha1, unknown_directory):
    dir_url = reverse("browse-directory", url_args={"sha1_git": invalid_sha1})

    check_html_get_response(
        client, dir_url, status_code=400, template_used="error.html"
    )

    dir_url = reverse("browse-directory", url_args={"sha1_git": unknown_directory})

    check_html_get_response(
        client, dir_url, status_code=404, template_used="error.html"
    )


def test_directory_masked(client, mocker, directory, make_masked_object_exception):
    masked_object_exception = make_masked_object_exception(f"swh:1:dir:{directory}")

    mock_archive = mocker.patch("swh.web.browse.utils.archive")
    mock_archive.lookup_directory.side_effect = masked_object_exception

    url = reverse("browse-directory", url_args={"sha1_git": directory})

    check_html_get_response(client, url, status_code=403, template_used="masked.html")
    check_http_get_response(
        client, url, content_type="application/json", status_code=403
    )
    check_http_get_response(
        client, url, content_type="application/yaml", status_code=403
    )


def test_directory_with_invalid_path(client, directory):
    path = "foo/bar"
    dir_url = reverse(
        "browse-directory",
        url_args={"sha1_git": directory},
        query_params={"path": path},
    )

    resp = check_html_get_response(
        client, dir_url, status_code=404, template_used="browse-directory.html"
    )
    error_message = (
        f"Directory entry with path {path} from root directory {directory} not found"
    )
    assert_contains(resp, error_message, status_code=404)


def test_directory_uppercase(client, directory):
    url = reverse(
        "browse-directory-uppercase-checksum", url_args={"sha1_git": directory.upper()}
    )

    resp = check_html_get_response(client, url, status_code=302)

    redirect_url = reverse("browse-directory", url_args={"sha1_git": directory})

    assert resp["location"] == redirect_url


def test_permalink_box_context(client, tests_data, directory):
    origin_url = random.choice(tests_data["origins"])["url"]
    url = reverse(
        "browse-directory",
        url_args={"sha1_git": directory},
        query_params={"origin_url": origin_url},
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-directory.html"
    )
    assert_contains(resp, 'id="swhid-context-option-directory"')


def test_directory_origin_snapshot_branch_browse(
    client, archive_data, origin_with_multiple_visits
):
    origin_url = origin_with_multiple_visits["url"]
    visits = archive_data.origin_visit_get(origin_url)
    visit = random.choice(visits)
    snapshot = archive_data.snapshot_get(visit["snapshot"])
    snapshot_sizes = archive_data.snapshot_count_branches(visit["snapshot"])
    branches, releases, _ = process_snapshot_branches(snapshot)
    branch_info = next(
        branch for branch in branches if branch["name"] == "refs/heads/master"
    )

    directory = archive_data.revision_get(branch_info["target"])["directory"]
    directory_content = archive_data.directory_ls(directory)
    directory_subdir = random.choice(
        [e for e in directory_content if e["type"] == "dir"]
    )

    url = reverse(
        "browse-directory",
        url_args={"sha1_git": directory},
        query_params={
            "origin_url": origin_url,
            "snapshot": snapshot["id"],
            "branch": branch_info["name"],
            "path": directory_subdir["name"],
        },
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-directory.html"
    )

    _check_origin_snapshot_related_html(
        resp, origin_with_multiple_visits, snapshot, snapshot_sizes, branches, releases
    )
    assert_contains(resp, directory_subdir["name"])
    assert_contains(resp, f"Branch: <strong>{branch_info['name']}</strong>")

    dir_swhid = gen_swhid(
        ObjectType.DIRECTORY,
        directory_subdir["target"],
        metadata={
            "origin": origin_url,
            "visit": gen_swhid(ObjectType.SNAPSHOT, snapshot["id"]),
            "anchor": gen_swhid(ObjectType.REVISION, branch_info["target"]),
            "path": "/",
        },
    )
    assert_contains(resp, dir_swhid)

    rev_swhid = gen_swhid(
        ObjectType.REVISION,
        branch_info["target"],
        metadata={
            "origin": origin_url,
            "visit": gen_swhid(ObjectType.SNAPSHOT, snapshot["id"]),
        },
    )
    assert_contains(resp, rev_swhid)

    snp_swhid = gen_swhid(
        ObjectType.SNAPSHOT,
        snapshot["id"],
        metadata={
            "origin": origin_url,
        },
    )
    assert_contains(resp, snp_swhid)


def test_directory_origin_snapshot_release_browse(
    client, archive_data, origin_with_multiple_visits
):
    origin_url = origin_with_multiple_visits["url"]
    visits = archive_data.origin_visit_get(origin_url)
    visit = random.choice(visits)
    snapshot = archive_data.snapshot_get(visit["snapshot"])
    snapshot_sizes = archive_data.snapshot_count_branches(visit["snapshot"])
    branches, releases, _ = process_snapshot_branches(snapshot)
    release_info = random.choice(releases)

    directory = release_info["directory"]
    directory_content = archive_data.directory_ls(directory)
    directory_subdir = random.choice(
        [e for e in directory_content if e["type"] == "dir"]
    )

    url = reverse(
        "browse-directory",
        url_args={"sha1_git": directory},
        query_params={
            "origin_url": origin_url,
            "snapshot": snapshot["id"],
            "release": release_info["name"],
            "path": directory_subdir["name"],
        },
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-directory.html"
    )

    _check_origin_snapshot_related_html(
        resp, origin_with_multiple_visits, snapshot, snapshot_sizes, branches, releases
    )
    assert_contains(resp, directory_subdir["name"])
    assert_contains(resp, f"Release: <strong>{release_info['name']}</strong>")

    dir_swhid = gen_swhid(
        ObjectType.DIRECTORY,
        directory_subdir["target"],
        metadata={
            "origin": origin_url,
            "visit": gen_swhid(ObjectType.SNAPSHOT, snapshot["id"]),
            "anchor": gen_swhid(ObjectType.RELEASE, release_info["id"]),
            "path": "/",
        },
    )
    assert_contains(resp, dir_swhid)

    rev_swhid = gen_swhid(
        ObjectType.REVISION,
        release_info["target"],
        metadata={
            "origin": origin_url,
            "visit": gen_swhid(ObjectType.SNAPSHOT, snapshot["id"]),
        },
    )
    assert_contains(resp, rev_swhid)

    rel_swhid = gen_swhid(
        ObjectType.RELEASE,
        release_info["id"],
        metadata={
            "origin": origin_url,
            "visit": gen_swhid(ObjectType.SNAPSHOT, snapshot["id"]),
        },
    )
    assert_contains(resp, rel_swhid)

    snp_swhid = gen_swhid(
        ObjectType.SNAPSHOT,
        snapshot["id"],
        metadata={
            "origin": origin_url,
        },
    )
    assert_contains(resp, snp_swhid)


def test_directory_origin_snapshot_revision_browse(
    client, archive_data, origin_with_multiple_visits
):
    origin_url = origin_with_multiple_visits["url"]
    visits = archive_data.origin_visit_get(origin_url)
    visit = random.choice(visits)
    snapshot = archive_data.snapshot_get(visit["snapshot"])
    branches, releases, _ = process_snapshot_branches(snapshot)
    branch_info = next(
        branch for branch in branches if branch["name"] == "refs/heads/master"
    )

    directory = archive_data.revision_get(branch_info["target"])["directory"]
    directory_content = archive_data.directory_ls(directory)
    directory_subdir = random.choice(
        [e for e in directory_content if e["type"] == "dir"]
    )

    url = reverse(
        "browse-directory",
        url_args={"sha1_git": directory},
        query_params={
            "origin_url": origin_url,
            "snapshot": snapshot["id"],
            "revision": branch_info["target"],
            "path": directory_subdir["name"],
        },
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-directory.html"
    )

    assert_contains(resp, f"Revision: <strong>{branch_info['target']}</strong>")


def _check_origin_snapshot_related_html(
    resp, origin, snapshot, snapshot_sizes, branches, releases
):
    browse_origin_url = reverse(
        "browse-origin", query_params={"origin_url": origin["url"]}
    )

    assert_contains(resp, f'href="{browse_origin_url}"')

    origin_branches_url = reverse(
        "browse-origin-branches",
        query_params={"origin_url": origin["url"], "snapshot": snapshot["id"]},
    )

    assert_contains(resp, f'href="{escape(origin_branches_url)}"')
    assert_contains(resp, f"Branches ({snapshot_sizes['revision']})")

    origin_releases_url = reverse(
        "browse-origin-releases",
        query_params={"origin_url": origin["url"], "snapshot": snapshot["id"]},
    )

    assert_contains(resp, f'href="{escape(origin_releases_url)}"')
    assert_contains(resp, f"Releases ({snapshot_sizes['release']})")

    assert_contains(resp, '<li class="swh-branch"', count=len(branches))
    assert_contains(resp, '<li class="swh-release"', count=len(releases))


def _directory_view_checks(
    client,
    root_directory_sha1,
    directory_entries,
    path=None,
    origin_url=None,
    snapshot_id=None,
    revision_id=None,
):
    dirs = [e for e in directory_entries if e["type"] in ("dir", "rev")]
    files = [e for e in directory_entries if e["type"] == "file"]

    url_args = {"sha1_git": root_directory_sha1}
    query_params = {"origin_url": origin_url, "snapshot": snapshot_id}

    url = reverse(
        "browse-directory",
        url_args=url_args,
        query_params={**query_params, "path": path},
    )

    root_dir_url = reverse(
        "browse-directory",
        url_args=url_args,
        query_params=query_params,
    )

    resp = check_html_get_response(
        client, url, status_code=200, template_used="browse-directory.html"
    )

    assert_contains(
        resp,
        '<a href="' + root_dir_url + '">' + root_directory_sha1[:7] + "</a>",
    )
    assert_contains(resp, '<td class="swh-directory">', count=len(dirs))
    assert_contains(resp, '<td class="swh-content">', count=len(files))

    for d in dirs:
        if d["type"] == "rev":
            dir_url = reverse("browse-revision", url_args={"sha1_git": d["target"]})
        else:
            dir_path = d["name"]
            if path:
                dir_path = "%s/%s" % (path, d["name"])
            dir_url = reverse(
                "browse-directory",
                url_args={"sha1_git": root_directory_sha1},
                query_params={**query_params, "path": dir_path},
            )
        assert_contains(resp, dir_url)

    for f in files:
        file_path = "%s/%s" % (root_directory_sha1, f["name"])
        if path:
            file_path = "%s/%s/%s" % (root_directory_sha1, path, f["name"])
        query_string = "sha1_git:" + f["target"]
        file_url = reverse(
            "browse-content",
            url_args={"query_string": query_string},
            query_params={**query_params, "path": file_path},
        )
        assert_contains(resp, file_url)

    path_info = gen_path_info(path)

    assert_contains(resp, '<li class="swh-path">', count=len(path_info) + 1)
    assert_contains(
        resp, '<a href="%s">%s</a>' % (root_dir_url, root_directory_sha1[:7])
    )

    for p in path_info:
        dir_url = reverse(
            "browse-directory",
            url_args={"sha1_git": root_directory_sha1},
            query_params={**query_params, "path": p["path"]},
        )
        assert_contains(resp, '<a href="%s">%s</a>' % (dir_url, p["name"]))

    assert_contains(resp, "vault-cook-directory")

    swh_dir_id = gen_swhid(ObjectType.DIRECTORY, directory_entries[0]["dir_id"])
    swh_dir_id_url = reverse("browse-swhid", url_args={"swhid": swh_dir_id})

    swhid_context = {}
    if origin_url:
        swhid_context["origin"] = origin_url
    if snapshot_id:
        swhid_context["visit"] = gen_swhid(ObjectType.SNAPSHOT, snapshot_id)
    if root_directory_sha1 != directory_entries[0]["dir_id"]:
        swhid_context["anchor"] = gen_swhid(ObjectType.DIRECTORY, root_directory_sha1)
    if root_directory_sha1 != directory_entries[0]["dir_id"]:
        swhid_context["anchor"] = gen_swhid(ObjectType.DIRECTORY, root_directory_sha1)
    if revision_id:
        swhid_context["anchor"] = gen_swhid(ObjectType.REVISION, revision_id)
    swhid_context["path"] = f"/{path}/" if path else None

    swh_dir_id = gen_swhid(
        ObjectType.DIRECTORY, directory_entries[0]["dir_id"], metadata=swhid_context
    )
    swh_dir_id_url = reverse("browse-swhid", url_args={"swhid": swh_dir_id})
    assert_contains(resp, swh_dir_id)
    assert_contains(resp, swh_dir_id_url)

    assert_not_contains(resp, "swh-metadata-popover")


@pytest.mark.django_db
@pytest.mark.parametrize("staff_user_logged_in", [False, True])
def test_browse_directory_snapshot_context_release_directory_target(
    client, staff_user, archive_data, directory_with_subdirs, staff_user_logged_in
):
    if staff_user_logged_in:
        client.force_login(staff_user)

    release_name = "v1.0.0"
    release = Release(
        name=release_name.encode(),
        message=f"release {release_name}".encode(),
        target=hash_to_bytes(directory_with_subdirs),
        target_type=ReleaseTargetType.DIRECTORY,
        synthetic=True,
    )
    archive_data.release_add([release])

    snapshot = Snapshot(
        branches={
            release_name.encode(): SnapshotBranch(
                target=release.id, target_type=SnapshotTargetType.RELEASE
            ),
        },
    )
    archive_data.snapshot_add([snapshot])

    dir_content = archive_data.directory_ls(directory_with_subdirs)
    dir_entry = random.choice(
        [entry for entry in dir_content if entry["type"] == "dir"]
    )

    browse_url = reverse(
        "browse-directory",
        url_args={"sha1_git": directory_with_subdirs},
        query_params={
            "path": dir_entry["name"],
            "release": release_name,
            "snapshot": snapshot.id.hex(),
        },
    )

    check_html_get_response(
        client, browse_url, status_code=200, template_used="browse-directory.html"
    )


@pytest.mark.parametrize("legacy_url", [True, False])
def test_browse_directory_with_path_targeting_file(
    client, archive_data, directory_with_files, legacy_url
):
    dir_content = archive_data.directory_ls(directory_with_files)
    file_entry = random.choice(
        [entry for entry in dir_content if entry["type"] == "file"]
    )

    if legacy_url:
        browse_url = reverse(
            "browse-directory-legacy",
            url_args={
                "sha1_git": directory_with_files,
                "path": file_entry["name"],
            },
        )
    else:
        browse_url = reverse(
            "browse-directory",
            url_args={"sha1_git": directory_with_files},
            query_params={
                "path": file_entry["name"],
            },
        )

    resp = check_http_get_response(client, browse_url, status_code=302)

    assert resp["location"] == reverse(
        "browse-content",
        url_args={"query_string": f"sha1_git:{file_entry['checksums']['sha1_git']}"},
        query_params={
            "path": f'{directory_with_files}/{file_entry["name"]}',
        },
    )

    check_http_get_response(client, resp["location"], status_code=200)


@given(new_person(), new_swh_date())
def test_directory_view_missing_readme_bytes(client, archive_data, person, date):
    origin_url = "https://git.example.org/user/project"
    root_dir = Directory(
        entries=(
            DirectoryEntry(
                name=b"README.md",
                type="file",
                # dummy sha1_git not available in objstorage
                target=hash_to_bytes("1" * 40),
                perms=DentryPerms.content,
            ),
        )
    )
    archive_data.directory_add([root_dir])

    revision = Revision(
        directory=root_dir.id,
        author=person,
        committer=person,
        message=b"commit message",
        date=TimestampWithTimezone.from_datetime(date),
        committer_date=TimestampWithTimezone.from_datetime(date),
        synthetic=False,
        type=RevisionType.GIT,
    )
    archive_data.revision_add([revision])

    snapshot = Snapshot(
        branches={
            b"HEAD": SnapshotBranch(
                target="refs/head/master".encode(),
                target_type=SnapshotTargetType.ALIAS,
            ),
            b"refs/head/master": SnapshotBranch(
                target=revision.id,
                target_type=SnapshotTargetType.REVISION,
            ),
        }
    )
    archive_data.snapshot_add([snapshot])

    archive_data.origin_add([Origin(url=origin_url)])
    date = now()
    visit = OriginVisit(origin=origin_url, date=date, type="git")
    visit = archive_data.origin_visit_add([visit])[0]
    visit_status = OriginVisitStatus(
        origin=origin_url,
        visit=visit.visit,
        date=date,
        status="full",
        snapshot=snapshot.id,
    )
    archive_data.origin_visit_status_add([visit_status])

    # check all views that render README files
    browse_dir_url = reverse(
        "browse-directory", url_args={"sha1_git": root_dir.id.hex()}
    )
    resp = check_html_get_response(
        client, browse_dir_url, status_code=200, template_used="browse-directory.html"
    )
    assert_contains(resp, "Readme bytes are not available")

    browse_rev_url = reverse(
        "browse-revision", url_args={"sha1_git": revision.id.hex()}
    )
    resp = check_html_get_response(
        client, browse_rev_url, status_code=200, template_used="browse-revision.html"
    )
    assert_contains(resp, "Readme bytes are not available")

    browse_origin_url = reverse(
        "browse-origin-directory", query_params={"origin_url": origin_url}
    )
    resp = check_html_get_response(
        client,
        browse_origin_url,
        status_code=200,
        template_used="browse-directory.html",
    )
    assert_contains(resp, "Readme bytes are not available")


def test_directory_notfound_xss(client):
    dir_url = reverse(
        "browse-origin-directory",
        query_params={"origin_url": "<script>alert(1)</script>"},
    )

    resp = check_html_get_response(
        client, dir_url, status_code=404, template_used="error.html"
    )

    assert_not_contains(resp, "<script>alert(1)</script>", status_code=404)
    assert_contains(resp, "&lt;script&gt;", status_code=404)



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import django.contrib.auth.models
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("auth", "0011_update_proxy_permissions"),
    ]

    operations = [
        migrations.CreateModel(
            name="OIDCUserOfflineTokens",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("user_id", models.CharField(max_length=50)),
                ("creation_date", models.DateTimeField(auto_now_add=True)),
                ("offline_token", models.BinaryField()),
            ],
            options={
                "db_table": "oidc_user_offline_tokens",
            },
        ),
        migrations.CreateModel(
            name="OIDCUser",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("auth.user",),
            managers=[
                ("objects", django.contrib.auth.models.UserManager()),
            ],
        ),
    ]



# Copyright (C) 2017-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import deque
import dataclasses
import hashlib
import json
import textwrap
from typing import Any, Dict, List, Optional

from django.http import HttpRequest, HttpResponse, JsonResponse
from django.shortcuts import render
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from swh.model.hashutil import hash_to_bytes
from swh.model.swhids import CoreSWHID, ObjectType
from swh.storage.algos.revisions_walker import State
from swh.web.browse.browseurls import browse_route
from swh.web.browse.snapshot_context import get_snapshot_context
from swh.web.browse.utils import (
    content_display_max_size,
    format_log_entries,
    gen_link,
    gen_person_mail_link,
    gen_revision_url,
    get_directory_entries,
    get_readme_to_display,
    get_revision_log_url,
    is_textual_content,
    prepare_content_for_display,
    pygments_iframe_height_for_content,
    request_content,
)
from swh.web.utils import (
    archive,
    cache_get,
    cache_set,
    format_utc_iso_date,
    gen_path_info,
    highlightjs,
    reverse,
    swh_object_icons,
)
from swh.web.utils.exc import NotFoundExc, http_status_code_message
from swh.web.utils.identifiers import get_swhids_info
from swh.web.utils.typing import RevisionMetadata, SnapshotContext, SWHObjectInfo


def _gen_content_url(
    revision: Dict[str, Any],
    query_string: str,
    path: str,
    snapshot_context: Optional[SnapshotContext],
) -> str:
    if snapshot_context:
        query_params = snapshot_context["query_params"]
        query_params["path"] = path
        query_params["revision"] = revision["id"]
        content_url = reverse("browse-origin-content", query_params=query_params)
    else:
        content_path = "%s/%s" % (revision["directory"], path)
        content_url = reverse(
            "browse-content",
            url_args={"query_string": query_string},
            query_params={"path": content_path},
        )
    return content_url


def _gen_diff_link(idx: int, diff_anchor: str, link_text: str) -> str:
    if idx < _max_displayed_file_diffs:
        return gen_link(diff_anchor, link_text)
    else:
        return link_text


# TODO: put in conf
_max_displayed_file_diffs = 1000


def _gen_revision_changes_list(
    revision: Dict[str, Any],
    changes: List[Dict[str, Any]],
    snapshot_context: Optional[SnapshotContext],
) -> str:
    """
    Returns a HTML string describing the file changes
    introduced in a revision.
    As this string will be displayed in the browse revision view,
    links to adequate file diffs are also generated.

    Args:
        revision (str): hexadecimal representation of a revision identifier
        changes (list): list of file changes in the revision
        snapshot_context (dict): optional origin context used to reverse
            the content urls

    Returns:
        A string to insert in a revision HTML view.

    """
    changes_msg = []
    for i, change in enumerate(changes):
        hasher = hashlib.sha1()
        from_query_string = ""
        to_query_string = ""
        diff_id = "diff-"
        if change["from"]:
            from_query_string = "sha1_git:" + change["from"]["target"]
            diff_id += change["from"]["target"] + "-" + change["from_path"]
        diff_id += "-"
        if change["to"]:
            to_query_string = "sha1_git:" + change["to"]["target"]
            diff_id += change["to"]["target"] + change["to_path"]
        change["path"] = change["to_path"] or change["from_path"]
        url_args = {
            "from_query_string": from_query_string,
            "to_query_string": to_query_string,
        }
        query_params = {"path": change["path"]}
        change["diff_url"] = reverse(
            "diff-contents", url_args=url_args, query_params=query_params
        )

        hasher.update(diff_id.encode("utf-8"))
        diff_id = hasher.hexdigest()
        change["id"] = diff_id
        diff_link = "#diff_" + diff_id

        if change["type"] == "modify":
            change["content_url"] = _gen_content_url(
                revision, to_query_string, change["to_path"], snapshot_context
            )
            changes_msg.append(
                "modified:  %s" % _gen_diff_link(i, diff_link, change["to_path"])
            )
        elif change["type"] == "insert":
            change["content_url"] = _gen_content_url(
                revision, to_query_string, change["to_path"], snapshot_context
            )
            changes_msg.append(
                "new file:  %s" % _gen_diff_link(i, diff_link, change["to_path"])
            )
        elif change["type"] == "delete":
            parent = archive.lookup_revision(revision["parents"][0])
            change["content_url"] = _gen_content_url(
                parent, from_query_string, change["from_path"], snapshot_context
            )
            changes_msg.append(
                "deleted:   %s" % _gen_diff_link(i, diff_link, change["from_path"])
            )
        elif change["type"] == "rename":
            change["content_url"] = _gen_content_url(
                revision, to_query_string, change["to_path"], snapshot_context
            )
            link_text = change["from_path"] + " &rarr; " + change["to_path"]
            changes_msg.append(
                "renamed:   %s" % _gen_diff_link(i, diff_link, link_text)
            )
    if not changes:
        changes_msg.append("No changes")
    return mark_safe("\n".join(changes_msg))


@browse_route(
    r"revision/(?P<sha1_git>[0-9a-f]+)/diff/",
    view_name="diff-revision",
    checksum_args=["sha1_git"],
)
def _revision_diff(request: HttpRequest, sha1_git: str) -> HttpResponse:
    """
    Browse internal endpoint to compute revision diff
    """
    revision = archive.lookup_revision(sha1_git)
    snapshot_context = None
    origin_url = request.GET.get("origin_url", None)
    if not origin_url:
        origin_url = request.GET.get("origin", None)
    timestamp = request.GET.get("timestamp", None)
    visit_id_str = request.GET.get("visit_id", None)
    visit_id = int(visit_id_str) if visit_id_str is not None else None
    if origin_url:
        snapshot_context = get_snapshot_context(
            origin_url=origin_url,
            timestamp=timestamp,
            visit_id=visit_id,
            visit_type=request.GET.get("visit_type"),
        )

    changes = archive.diff_revision(sha1_git)
    changes_msg = _gen_revision_changes_list(revision, changes, snapshot_context)

    diff_data = {
        "total_nb_changes": len(changes),
        "changes": changes[:_max_displayed_file_diffs],
        "changes_msg": changes_msg,
    }
    return JsonResponse(diff_data)


NB_LOG_ENTRIES = 100


@browse_route(
    r"revision/(?P<sha1_git>[0-9a-f]+)/log/",
    view_name="browse-revision-log",
    checksum_args=["sha1_git"],
)
def revision_log_browse(request: HttpRequest, sha1_git: str) -> HttpResponse:
    """
    Django view that produces an HTML display of the history
    log for a revision identified by its id.

    The url that points to it is :http:get:`/browse/revision/(sha1_git)/log/`
    """
    origin_url = request.GET.get("origin_url")
    snapshot_id = request.GET.get("snapshot")
    snapshot_context = None
    if origin_url or snapshot_id:
        visit_id = int(request.GET.get("visit_id", 0))
        snapshot_context = get_snapshot_context(
            snapshot_id=snapshot_id,
            origin_url=origin_url,
            timestamp=request.GET.get("timestamp"),
            visit_id=visit_id or None,
            branch_name=request.GET.get("branch"),
            release_name=request.GET.get("release"),
            revision_id=sha1_git,
            visit_type=request.GET.get("visit_type"),
        )
    per_page = int(request.GET.get("per_page", NB_LOG_ENTRIES))
    offset = int(request.GET.get("offset", 0))
    revs_ordering = request.GET.get("revs_ordering", "committer_date")
    cache_key = f"rev_{sha1_git}_log_ordering_{revs_ordering}"
    rev_log_session = cache_get(
        cache_key,
        extra_decoders={
            "rev_walker_state": lambda d: State(
                done=d["done"],
                last_rev=d["last_rev"],
                num_revs=d["num_revs"],
                missing_revs=d["missing_revs"],
                revs_to_visit=(
                    list(map(tuple, d["revs_to_visit"]))
                    if revs_ordering == "committer_date"
                    else d["revs_to_visit"]
                ),
            ),
            "set": set,
            "deque": deque,
        },
    )

    rev_log = []
    revs_walker_state = None
    if rev_log_session:
        rev_log = rev_log_session["rev_log"]
        revs_walker_state = rev_log_session["revs_walker_state"]

    if len(rev_log) < offset + per_page:
        revs_walker = archive.get_revisions_walker(
            revs_ordering,
            sha1_git,
            max_revs=offset + per_page + 1,
            state=revs_walker_state,
        )

        rev_log += [rev["id"] for rev in revs_walker]
        revs_walker_state = revs_walker.export_state()
        cache_set(
            cache_key,
            {
                "rev_log": rev_log,
                "revs_walker_state": revs_walker_state,
            },
            timeout=60 * 60,  # one hour
            extra_encoders=[
                (State, "rev_walker_state", dataclasses.asdict),
                (set, "set", list),
                (deque, "deque", list),
            ],
        )

    revs = rev_log[offset : offset + per_page]
    revision_log = archive.lookup_revision_multiple(revs)

    revs_ordering = request.GET.get("revs_ordering", "")

    prev_log_url = None
    if len(rev_log) > offset + per_page:
        prev_log_url = reverse(
            "browse-revision-log",
            url_args={"sha1_git": sha1_git},
            query_params={
                "per_page": str(per_page),
                "offset": str(offset + per_page),
                "revs_ordering": revs_ordering or None,
                "origin_url": origin_url,
                "snapshot": snapshot_id,
            },
        )

    next_log_url = None
    if offset != 0:
        next_log_url = reverse(
            "browse-revision-log",
            url_args={"sha1_git": sha1_git},
            query_params={
                "per_page": str(per_page),
                "offset": str(offset - per_page),
                "revs_ordering": revs_ordering or None,
                "origin_url": origin_url,
                "snapshot": snapshot_id,
            },
        )

    revision_log_data = format_log_entries(revision_log, per_page, snapshot_context)

    swh_rev_id = str(
        CoreSWHID(object_type=ObjectType.REVISION, object_id=hash_to_bytes(sha1_git))
    )

    return render(
        request,
        "browse-revision-log.html",
        {
            "heading": "Revision history",
            "swh_object_id": swh_rev_id,
            "swh_object_name": "Revisions history",
            "swh_object_metadata": None,
            "revision_log": revision_log_data,
            "revs_ordering": revs_ordering,
            "next_log_url": next_log_url,
            "prev_log_url": prev_log_url,
            "breadcrumbs": None,
            "top_right_link": None,
            "snapshot_context": snapshot_context,
            "vault_cooking": None,
            "show_actions": True,
            "swhids_info": None,
        },
    )


@browse_route(
    r"revision/(?P<sha1_git>[0-9a-f]+)/",
    view_name="browse-revision",
    checksum_args=["sha1_git"],
)
def revision_browse(request: HttpRequest, sha1_git: str) -> HttpResponse:
    """
    Django view that produces an HTML display of a revision
    identified by its id.

    The url that points to it is :http:get:`/browse/revision/(sha1_git)/`.
    """
    revision = archive.lookup_revision(sha1_git)
    origin_info = None
    snapshot_context = None
    origin_url = request.GET.get("origin_url")
    if not origin_url:
        origin_url = request.GET.get("origin")
    timestamp = request.GET.get("timestamp")
    visit_id = int(request.GET.get("visit_id", 0))
    snapshot_id = request.GET.get("snapshot_id")
    if not snapshot_id:
        snapshot_id = request.GET.get("snapshot")
    path = request.GET.get("path")
    dir_id = None
    dirs, files = [], []
    content_data = {}
    if origin_url:
        try:
            snapshot_context = get_snapshot_context(
                snapshot_id=snapshot_id,
                origin_url=origin_url,
                timestamp=timestamp,
                visit_id=visit_id or None,
                branch_name=request.GET.get("branch"),
                release_name=request.GET.get("release"),
                revision_id=sha1_git,
                path=path,
                visit_type=request.GET.get("visit_type"),
            )
        except NotFoundExc as e:
            raw_rev_url = reverse(
                "browse-revision", url_args={"sha1_git": sha1_git}, request=request
            )
            error_message = format_html(
                "The Software Heritage archive has a revision "
                "with the hash you provided but the origin "
                "mentioned in your request appears broken: {}. "
                "Please check the URL and try again.\n\n"
                "Nevertheless, you can still browse the revision "
                "without origin information: {}",
                origin_url,
                raw_rev_url,
            )
            if str(e).startswith("Origin"):
                raise NotFoundExc(error_message)
            else:
                raise e
        origin_info = snapshot_context["origin_info"]
        snapshot_id = snapshot_context["snapshot_id"]
    elif snapshot_id:
        snapshot_context = get_snapshot_context(snapshot_id)

    error_info: Dict[str, Any] = {"status_code": 200, "description": None}
    extra_template_variables = {}

    if path:
        try:
            file_info = archive.lookup_directory_with_path(revision["directory"], path)
            if file_info["type"] == "dir":
                dir_id = file_info["target"]
            else:
                query_string = "sha1_git:" + file_info["target"]
                content_data = request_content(query_string)
                extra_template_variables["sha1_git"] = file_info["target"]
        except NotFoundExc as e:
            error_info["status_code"] = 404
            error_info["description"] = f"NotFoundExc: {str(e)}"
    else:
        dir_id = revision["directory"]

    if dir_id:
        path = "" if not path else (path + "/")
        dirs, files = get_directory_entries(dir_id)

    revision_metadata = RevisionMetadata(
        object_type=ObjectType.REVISION,
        object_id=sha1_git,
        revision=sha1_git,
        author=revision["author"]["fullname"] if revision["author"] else "None",
        author_url=(
            gen_person_mail_link(revision["author"]) if revision["author"] else "None"
        ),
        committer=(
            revision["committer"]["fullname"] if revision["committer"] else "None"
        ),
        committer_url=(
            gen_person_mail_link(revision["committer"])
            if revision["committer"]
            else "None"
        ),
        committer_date=format_utc_iso_date(revision["committer_date"]),
        date=format_utc_iso_date(revision["date"]),
        directory=revision["directory"],
        merge=revision["merge"],
        metadata=json.dumps(
            revision["metadata"], sort_keys=True, indent=4, separators=(",", ": ")
        ),
        parents=revision["parents"],
        synthetic=revision["synthetic"],
        type=revision["type"],
        snapshot=snapshot_id,
        origin_url=origin_url,
        root_directory=snapshot_context["root_directory"] if snapshot_context else None,
    )

    message_lines = ["None"]
    if revision["message"]:
        message_lines = revision["message"].split("\n")

    parents = []
    for p in revision["parents"]:
        parent_url = gen_revision_url(p, snapshot_context)
        parents.append({"id": p, "url": parent_url})

    path_info = gen_path_info(path)

    query_params = snapshot_context["query_params"] if snapshot_context else {}

    breadcrumbs = []
    breadcrumbs.append(
        {
            "name": revision["directory"][:7],
            "url": reverse(
                "browse-revision",
                url_args={"sha1_git": sha1_git},
                query_params={**query_params, **{"path": ""}},
            ),
        }
    )
    for pi in path_info:
        query_params["path"] = pi["path"]
        breadcrumbs.append(
            {
                "name": pi["name"],
                "url": reverse(
                    "browse-revision",
                    url_args={"sha1_git": sha1_git},
                    query_params=query_params,
                ),
            }
        )

    vault_cooking = {
        "content_context": bool(content_data),
        "directory_context": not bool(content_data),
        "directory_swhid": f"swh:1:dir:{revision['directory']}",
        "revision_context": not bool(content_data),
        "revision_swhid": f"swh:1:rev:{sha1_git}",
    }

    swh_objects = [SWHObjectInfo(object_type=ObjectType.REVISION, object_id=sha1_git)]

    available_languages = None
    content = None
    content_size = None
    filename = None
    mimetype = None
    language = None
    readme_name = None
    readme_url = None
    readme_html = None
    readmes = {}
    textual_content = False

    extra_context = dict(revision_metadata)
    extra_context["path"] = f"/{path}" if path else None

    if content_data:
        breadcrumbs[-1]["url"] = None
        content_size = content_data["length"]
        mimetype = content_data["mimetype"]
        if content_data["raw_data"]:
            content_display_data = prepare_content_for_display(
                content_data["raw_data"],
                content_data["mimetype"],
                content_data["encoding"],
                path,
            )
            content = content_display_data["content_data"]
            language = content_display_data["language"]
            mimetype = content_display_data["mimetype"]
            if mimetype and is_textual_content(
                content_data["mimetype"], content_data["encoding"]
            ):
                available_languages = highlightjs.get_supported_languages()
                textual_content = True
            extra_template_variables["no_script_iframe_height"] = (
                pygments_iframe_height_for_content(content)
            )
        if path:
            filename = path_info[-1]["name"]
            query_params["filename"] = filename
            filepath = "/".join(pi["name"] for pi in path_info[:-1])
            extra_context["path"] = f"/{filepath}/" if filepath else "/"
            extra_context["filename"] = filename

            if filepath:
                dir_id = archive.lookup_directory_with_path(
                    revision["directory"], filepath
                )["target"]
            else:
                dir_id = revision["directory"]
            swh_objects.append(
                SWHObjectInfo(object_type=ObjectType.DIRECTORY, object_id=dir_id)
            )

        top_right_link = {
            "url": reverse(
                "browse-content-raw",
                url_args={"query_string": query_string},
                query_params={"filename": filename},
            ),
            "icon": swh_object_icons["content"],
            "text": "Raw File",
        }

        swh_objects.append(
            SWHObjectInfo(object_type=ObjectType.CONTENT, object_id=file_info["target"])
        )
        vault_cooking["content_download_url"] = reverse(
            "api-1-content-raw",
            url_args={"q": f"sha1_git:{file_info['target']}"},
            query_params={"filename": filename},
        )
    else:
        for d in dirs:
            if d["type"] == "rev":
                d["url"] = reverse(
                    "browse-revision", url_args={"sha1_git": d["target"]}
                )
            else:
                query_params["path"] = path + d["name"]
                d["url"] = reverse(
                    "browse-revision",
                    url_args={"sha1_git": sha1_git},
                    query_params=query_params,
                )
        for f in files:
            query_params["path"] = path + f["name"]
            f["url"] = reverse(
                "browse-revision",
                url_args={"sha1_git": sha1_git},
                query_params=query_params,
            )
            if f["name"].lower().startswith("readme"):
                readmes[f["name"]] = f.get("target")

        readme_name, readme_url, readme_html = get_readme_to_display(readmes)

        top_right_link = {
            "url": get_revision_log_url(sha1_git, snapshot_context),
            "icon": swh_object_icons["revisions history"],
            "text": "History",
        }

        vault_cooking["directory_context"] = True
        vault_cooking["directory_swhid"] = f"swh:1:dir:{dir_id}"

        swh_objects.append(
            SWHObjectInfo(object_type=ObjectType.DIRECTORY, object_id=dir_id)
        )

    query_params.pop("path", None)

    diff_revision_url = reverse(
        "diff-revision",
        url_args={"sha1_git": sha1_git},
        query_params=query_params,
    )

    if snapshot_id:
        swh_objects.append(
            SWHObjectInfo(object_type=ObjectType.SNAPSHOT, object_id=snapshot_id)
        )

    swhids_info = get_swhids_info(swh_objects, snapshot_context, extra_context)

    heading = "Revision - %s - %s" % (
        sha1_git[:7],
        textwrap.shorten(message_lines[0], width=70),
    )
    if snapshot_context:
        context_found = "snapshot: %s" % snapshot_context["snapshot_id"]
        if origin_info:
            context_found = "origin: %s" % origin_info["url"]
        heading += " - %s" % context_found

    return render(
        request,
        "browse-revision.html",
        {
            "heading": heading,
            "swh_object_id": swhids_info[0]["swhid"],
            "swh_object_name": "Revision",
            "swh_object_metadata": revision_metadata,
            "message_header": message_lines[0],
            "message_body": "\n".join(message_lines[1:]),
            "parents": parents,
            "snapshot_context": snapshot_context,
            "dirs": dirs,
            "files": files,
            "content": content,
            "content_size": content_size,
            "max_content_size": content_display_max_size,
            "filename": filename,
            "encoding": content_data.get("encoding"),
            "mimetype": mimetype,
            "language": language,
            "readme_name": readme_name,
            "readme_url": readme_url,
            "readme_html": readme_html,
            "breadcrumbs": breadcrumbs,
            "top_right_link": top_right_link,
            "vault_cooking": vault_cooking,
            "diff_revision_url": diff_revision_url,
            "show_actions": True,
            "swhids_info": swhids_info,
            "error_code": error_info["status_code"],
            "error_message": http_status_code_message.get(error_info["status_code"]),
            "error_description": error_info["description"],
            "available_languages": available_languages,
            "textual_content": textual_content,
            **extra_template_variables,
        },
        status=error_info["status_code"],
    )



# Copyright (C) 2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django_bootstrap5.renderers import FieldRenderer


class CustomValidationClassFieldRenderer(FieldRenderer):
    def get_server_side_validation_classes(self) -> str:
        """Return CSS classes for server-side validation.

        This overrides the default field renderer to properly use ``error_css_class``
        and ``success_css_class`` parameters which are otherwise ignored.

        @see: https://github.com/zostera/django-bootstrap5/issues/302

        Returns:
            A class name to apply to a field.
        """
        if self.field_errors:
            return self.error_css_class
        elif self.field.form.is_bound:
            return self.success_css_class
        return ""



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.urls import re_path as url

from swh.web.add_forge_now.admin_views import (
    add_forge_now_request_dashboard,
    add_forge_now_requests_moderation_dashboard,
)

# register Web API endpoints
from swh.web.add_forge_now.api_views import add_forge_now_api_urls
from swh.web.add_forge_now.views import (
    add_forge_request_list_datatables,
    create_request_create,
    create_request_help,
    create_request_list,
    create_request_message_source,
)

urlpatterns = [
    url(
        r"^add-forge/request/list/datatables/$",
        add_forge_request_list_datatables,
        name="add-forge-request-list-datatables",
    ),
    url(r"^add-forge/request/create/$", create_request_create, name="forge-add-create"),
    url(r"^add-forge/request/list/$", create_request_list, name="forge-add-list"),
    url(
        r"^add-forge/request/message-source/(?P<id>\d+)/$",
        create_request_message_source,
        name="forge-add-message-source",
    ),
    url(r"^add-forge/request/help/$", create_request_help, name="forge-add-help"),
    url(
        r"^admin/add-forge/requests/$",
        add_forge_now_requests_moderation_dashboard,
        name="add-forge-now-requests-moderation",
    ),
    url(
        r"^admin/add-forge/request/(?P<request_id>(\d)+)/$",
        add_forge_now_request_dashboard,
        name="add-forge-now-request-dashboard",
    ),
] + add_forge_now_api_urls.get_url_patterns()



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from sentry_sdk.integrations.django import DjangoIntegration

from swh.core.sentry import init_sentry


def post_fork(server, worker):
    init_sentry(sentry_dsn=None, integrations=[DjangoIntegration()])



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from io import StringIO
from typing import Dict

import pytest

from django.core.management import call_command
from django.db import transaction

from swh.web.mailmap.models import UserMailmap

from .test_mailmap import populate_mailmap

pytest_plugins = [
    "swh.storage.tests.masking.conftest",
]


def call_sync_mailmaps(*args) -> str:
    out = StringIO()
    err = StringIO()

    call_command("sync_masking_mailmaps", *args, stdout=out, stderr=err)

    out.seek(0)
    err.seek(0)
    assert err.read() == ""
    return out.read()


def get_displaynames(masking_db) -> Dict[str, str]:
    with masking_db.cursor() as cur:
        cur.execute("SELECT original_email, display_name FROM display_name")

        return {bytes(f).decode("utf-8"): bytes(d).decode("utf-8") for (f, d) in cur}


@pytest.mark.django_db(transaction=True)
def test_sync_mailmaps_dry_run(masking_db_postgresql):
    with transaction.atomic():
        populate_mailmap()

    dsn = masking_db_postgresql.info.dsn

    out = call_sync_mailmaps(dsn)

    assert "(dry run)" in out
    assert "Synced 1 mailmaps to the masking proxy database" in out

    assert get_displaynames(masking_db_postgresql) == {}

    assert (
        UserMailmap.objects.filter(
            from_email_verified=True,
            display_name_activated=True,
            mailmap_last_processing_date__isnull=False,
        ).count()
        == 0
    )


@pytest.mark.django_db(transaction=True)
def test_sync_mailmaps_perform(masking_db_postgresql):
    with transaction.atomic():
        populate_mailmap()

    dsn = masking_db_postgresql.info.dsn

    out = call_sync_mailmaps("--perform", dsn)

    assert "(dry run)" not in out
    assert "Synced 1 mailmaps to the masking proxy database" in out

    expected_displaynames = {
        "from_emailVA@example.com": "Display Name V A",
    }

    assert get_displaynames(masking_db_postgresql) == expected_displaynames

    assert (
        UserMailmap.objects.filter(
            from_email_verified=True,
            display_name_activated=True,
            mailmap_last_processing_date__isnull=False,
        ).count()
        == 1
    )

    # do it again: nothing should be done
    out = call_sync_mailmaps("--perform", dsn)
    assert "No mailmap to update" in out
    assert get_displaynames(masking_db_postgresql) == expected_displaynames

    # do it again with --force
    out = call_sync_mailmaps("--perform", "--force", dsn)
    assert "No mailmap to update" not in out
    assert "(dry run)" not in out
    assert "Synced 1 mailmaps to the masking proxy database" in out
    assert get_displaynames(masking_db_postgresql) == expected_displaynames


@pytest.mark.django_db(transaction=True)
def test_sync_mailmaps_with_to_email(masking_db_postgresql):
    with transaction.atomic():
        populate_mailmap()

    dsn = masking_db_postgresql.info.dsn

    call_sync_mailmaps("--perform", dsn)

    expected_displaynames = {
        "from_emailVA@example.com": "Display Name V A",
    }

    assert get_displaynames(masking_db_postgresql) == expected_displaynames
    # Add a non-valid to_email
    with transaction.atomic():
        for mailmap in UserMailmap.objects.filter(
            from_email_verified=True, display_name_activated=True
        ):
            mailmap.to_email = "to_email@example.com"
            mailmap.save()

    call_sync_mailmaps("--perform", dsn)

    assert get_displaynames(masking_db_postgresql) == expected_displaynames

    # Verify the relevant to_email
    with transaction.atomic():
        for mailmap in UserMailmap.objects.filter(
            from_email_verified=True, display_name_activated=True
        ):
            mailmap.to_email_verified = True
            mailmap.save()

    call_sync_mailmaps("--perform", dsn)

    new_displayname = "Display Name V A <to_email@example.com>"
    expected_displaynames = {
        "from_emailVA@example.com": new_displayname,
    }

    assert get_displaynames(masking_db_postgresql) == expected_displaynames


@pytest.mark.django_db(transaction=True)
def test_sync_mailmaps_disable(masking_db_postgresql):
    with transaction.atomic():
        populate_mailmap()

    dsn = masking_db_postgresql.info.dsn

    # Do the initial mailmap sync
    call_sync_mailmaps("--perform", dsn)

    assert len(get_displaynames(masking_db_postgresql)) == 1

    updated = 0
    # Disable a display name
    with transaction.atomic():
        # Cannot use update() because `last_update_date` would not be updated
        for mailmap in UserMailmap.objects.filter(
            from_email_verified=True, display_name_activated=True
        ):
            mailmap.display_name_activated = False
            mailmap.save()
            updated += 1

    assert updated == 1

    # Sync mailmaps again
    out = call_sync_mailmaps("--perform", dsn)
    assert "1 mailmaps to disable" in out

    assert get_displaynames(masking_db_postgresql) == {}



# Copyright (C) 2024-2026  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information


from contextlib import contextmanager
from email.message import EmailMessage
import re
from typing import Iterator
from unittest.mock import MagicMock

import pytest

from django.dispatch import Signal

from swh.web.inbound_email import signals


@contextmanager
def mock_signal_receiver(
    signal: Signal,
    name: str = "receiver_name",
    disconnect_old_receivers: bool = True,
) -> Iterator[MagicMock]:
    receiver = MagicMock()
    receiver.configure_mock(__name__=name, __qualname__=name)

    try:
        if disconnect_old_receivers:
            old_receivers = signal.receivers
            signal.receivers = []
        signal.connect(receiver)
        yield receiver
    finally:
        signal.disconnect(receiver)
        if disconnect_old_receivers:
            signal.receivers = old_receivers


@pytest.fixture
def inbound_message(request) -> bytes:
    marker = request.node.get_closest_marker("inbound_message")
    if marker is None:
        raise ValueError("Missing inbound_message data")
    return marker.args[0]


def _test_message() -> bytes:
    message = EmailMessage()
    message["to"] = "test@example.com"
    message["subject"] = "Test Subject"
    message.set_content("This is a test message.\n")

    return bytes(message)


TEST_MESSAGE = _test_message()


@pytest.mark.inbound_message(TEST_MESSAGE)
@pytest.mark.parametrize(
    "return_value,err_contents",
    [
        # When the email gets processed by one of the receivers, the management command
        # should not emit any output.
        (signals.EmailProcessingStatus.PROCESSED, ""),
        # When a receiver fails, the management command outputs a message to this effect
        (signals.EmailProcessingStatus.FAILED, "Failed receiver.*receiver_name"),
        # When all receivers ignore a message, this fact is printed too
        (signals.EmailProcessingStatus.IGNORED, "Unhandled message"),
    ],
)
def test_signal_receiver(return_value, err_contents, caplog, receive_inbound_message):
    """Check that signal receivers are properly called when running the management command.

    Check for output depending on its return value"""

    with mock_signal_receiver(signals.email_received) as receiver:
        receiver.return_value = return_value

        message = receive_inbound_message(
            expect_failure=return_value != signals.EmailProcessingStatus.PROCESSED
        )

        output = "\n".join(record.getMessage() for record in caplog.records)

        if err_contents:
            assert re.match(err_contents, output)
        else:
            assert output == ""

        calls = receiver.call_args_list

        assert len(calls) == 1
        assert bytes(calls[0][1]["message"]) == message


@pytest.mark.inbound_message(TEST_MESSAGE)
def test_multiple_receivers(caplog, receive_inbound_message):
    with (
        mock_signal_receiver(signals.email_received, name="ignored") as ignored,
        mock_signal_receiver(
            signals.email_received, name="processed", disconnect_old_receivers=False
        ) as processed,
    ):
        ignored.return_value = signals.EmailProcessingStatus.IGNORED
        processed.return_value = signals.EmailProcessingStatus.PROCESSED

        message = receive_inbound_message()

        assert not caplog.records

        for receiver in [ignored, processed]:
            calls = receiver.call_args_list

            assert len(calls) == 1
            assert bytes(calls[0][1]["message"]) == message


@pytest.mark.inbound_message(TEST_MESSAGE)
def test_signal_receiver_exception(caplog, receive_inbound_message):
    with mock_signal_receiver(
        signals.email_received, name="exception_raised"
    ) as receiver:
        receiver.side_effect = ValueError("I'm broken!")

        receive_inbound_message(expect_failure=True)

        output = "\n".join(
            record.getMessage() + ("\n" + record.exc_text if record.exc_text else "")
            for record in caplog.records
        )

        assert re.search("Failed receiver.*exception_raised", output)
        assert "following exception" in output
        assert "ValueError" in output
        assert "I'm broken" in output



# Copyright (C) 2022-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from django.urls import re_path as url

from swh.web.banners.views import downtime_banner, fundraising_banner, hiring_banner

urlpatterns = [
    url(r"^fundraising/banner/$", fundraising_banner, name="swh-fundraising-banner"),
    url(r"^hiring/banner/$", hiring_banner, name="swh-hiring-banner"),
    url(r"^downtime/banner/$", downtime_banner, name="swh-downtime-banner"),
]



# Copyright (C) 2021-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

import re
from typing import Any, Dict

from django.http import HttpResponse
from django.shortcuts import redirect
from rest_framework import serializers
from rest_framework.request import Request

from swh.model import hashutil, swhids
from swh.model.model import MetadataAuthority, MetadataAuthorityType, Origin
from swh.web import config
from swh.web.api.apidoc import api_doc, format_docstring
from swh.web.api.apiurls import api_route
from swh.web.api.serializers import SoftLimitsIntegerField
from swh.web.utils import archive, reverse
from swh.web.utils.exc import BadInputExc, NotFoundExc


class MetadataAuthorityField(serializers.Field):
    """A DRF field to handle metadata authorities."""

    def to_representation(self, value: MetadataAuthority) -> str:
        """Serialize value.

        Args:
            value: a MetadataAuthority

        Returns:
            A metadata authority identifier, formatted as ``type IRI``
        """
        return f"{value.type.value} {value.url}"

    def to_internal_value(self, data: str) -> MetadataAuthority:
        """From ``type IRI`` to MetadataAuthority.

        Handles serialization and validation of a metadata authority.

        Args:
            data: A metadata authority identifier, formatted as ``type IRI``

        Raises:
            serializers.ValidationError: invalid value (missing space, invalid type).

        Returns:
            A MetadataAuthority
        """
        authority = data.strip()
        if " " not in authority:
            raise serializers.ValidationError(
                "The 'authority' query parameter must contain a space."
            )
        type_, url = authority.split(" ", 1)
        type_choices = [e.value for e in MetadataAuthorityType]
        if type_ not in type_choices:
            raise serializers.ValidationError(
                f"Invalid type {type_}, must be one of: {', '.join(type_choices)}"
            )
        return MetadataAuthority(MetadataAuthorityType(type_), url)


class RawExtrinsicMetadataQuerySerializer(serializers.Serializer):
    """Raw Extrinsic Metadata query parameters serializer."""

    authority = MetadataAuthorityField(required=True)
    after = serializers.DateTimeField(required=False, default=None)
    limit = SoftLimitsIntegerField(
        required=False, default=100, min_value=1, max_value=10000
    )
    page_token = serializers.CharField(required=False, default=None)


@api_route(
    "/raw-extrinsic-metadata/swhid/<swhid:target>/",
    "api-1-raw-extrinsic-metadata-swhid",
    query_params_serializer=RawExtrinsicMetadataQuerySerializer,
)
@api_doc("/raw-extrinsic-metadata/swhid/", category="Metadata")
@format_docstring()
def api_raw_extrinsic_metadata_swhid(
    request: Request,
    target: str,
    validated_query_params: dict[str, Any],
):
    """
    .. http:get:: /api/1/raw-extrinsic-metadata/swhid/(target)

        Returns raw `extrinsic metadata <https://docs.softwareheritage.org/devel/glossary.html#term-extrinsic-metadata>`__ collected on a given object.

        :param string target: The core SWHID of the object whose metadata
            should be returned
        :query string authority: A metadata authority identifier, formatted as
            ``<type> <IRI>``. Required.
        :query string after: ISO8601 representation of the minimum timestamp of metadata
            to fetch. Defaults to allowing all metadata.
        :query int limit: Maximum number of metadata objects to return (default to 100).
        :query string page_token: optional opaque token, used to get the next page of
            results

        {common_headers}

        :>jsonarr string target: SWHID of the object described by this metadata
            (absent when ``target`` is not a core SWHID (ie. it does not have type
            ``cnt``/``dir``/``rev``/``rel``/``snp``)
        :>jsonarr string discovery_date: ISO8601/RFC3339 timestamp of the moment this
            metadata was collected.
        :>jsonarr object authority: authority this metadata is coming from
        :>jsonarr object fetcher: tool used to fetch the metadata
        :>jsonarr string format: short identifier of the format of the metadata
        :>jsonarr string metadata_url: link to download the metadata "blob" itself
        :>jsonarr string origin: URL of the origin in which context
            the metadata is valid, if any
        :>jsonarr int visit: identifier of the visit in which context
            the metadata is valid, if any
        :>jsonarr string snapshot: SWHID of the snapshot in which context
            the metadata is valid, if any
        :>jsonarr string release: SWHID of the release in which context
            the metadata is valid, if any
        :>jsonarr string revision: SWHID of the revision in which context
            the metadata is valid, if any
        :>jsonarr string path: SWHID of the path in which context
            is valid if any, relative to a release or revision as anchor
        :>jsonarr string directory: SWHID of the directory in which context
            the metadata is valid, if any

        :statuscode 200: no error

        **Example:**

        .. parsed-literal::

            :swh_web_api:`raw-extrinsic-metadata/swhid/swh:1:dir:a2faa28028657859c16ff506924212b33f0e1307/?authority=forge%20https://pypi.org/`
    """  # noqa B950

    authority = validated_query_params["authority"]
    after = validated_query_params["after"]
    page_token = validated_query_params["page_token"]
    limit = validated_query_params["limit"]

    try:
        parsed_target = swhids.ExtendedSWHID.from_string(target)
    except swhids.ValidationError as e:
        raise BadInputExc(f"Invalid target SWHID: {e}") from None

    result_page = archive.lookup_raw_extrinsic_metadata(
        parsed_target, authority, after, page_token, limit
    )

    filename = None
    if parsed_target.object_type == swhids.ExtendedObjectType.ORIGIN:
        origin_sha1 = hashutil.hash_to_hex(parsed_target.object_id)
        (origin_info,) = list(archive.lookup_origins_by_sha1s([origin_sha1]))
        if origin_info is not None:
            filename = re.sub("[:/_.]+", "_", origin_info["url"]) + "_metadata"
    if filename is None:
        filename = f"{target}_metadata"

    results = []

    for metadata in result_page.results:
        # We can't reliably send metadata directly, because it is a bytestring,
        # and we have to return JSON documents.
        metadata["metadata_url"] = reverse(
            "api-1-raw-extrinsic-metadata-get",
            url_args={"id": metadata["id"]},
            query_params={"filename": filename},
            request=request,
        )

        del metadata["id"]
        results.append(metadata)

    headers: Dict[str, str] = {}
    if result_page.next_page_token is not None:
        headers["link-next"] = reverse(
            "api-1-raw-extrinsic-metadata-swhid",
            url_args={"target": target},
            query_params=dict(
                authority=request.query_params["authority"],
                after=str(after) if after else None,
                limit=limit,
                page_token=result_page.next_page_token,
            ),
            request=request,
        )

    return {
        "results": results,
        "headers": headers,
    }


@api_route(
    "/raw-extrinsic-metadata/get/(?P<id>[0-9a-z]+)/",
    "api-1-raw-extrinsic-metadata-get",
)
def api_raw_extrinsic_metadata_get(request: Request, id: str):
    # This is an internal endpoint that should only be accessed via URLs given
    # by /raw-extrinsic-metadata/swhid/; so it is not documented.
    metadata = config.storage().raw_extrinsic_metadata_get_by_ids(
        [hashutil.hash_to_bytes(id)]
    )
    if not metadata:
        raise NotFoundExc(
            "Metadata not found. Use /raw-extrinsic-metadata/swhid/ to access metadata."
        )

    response = HttpResponse(
        metadata[0].metadata, content_type="application/octet-stream"
    )

    filename = request.query_params.get("filename")
    if filename and re.match("[a-zA-Z0-9:._-]+", filename):
        response["Content-disposition"] = f'attachment; filename="{filename}"'
    else:
        # It should always be not-None and match the regexp if the URL was created by
        # /raw-extrinsic-metadata/swhid/, but we're better safe than sorry.
        response["Content-disposition"] = "attachment"

    return response


@api_route(
    "/raw-extrinsic-metadata/swhid/<swhid:target>/authorities/",
    "api-1-raw-extrinsic-metadata-swhid-authorities",
)
@api_doc("/raw-extrinsic-metadata/swhid/authorities/", category="Metadata")
@format_docstring()
def api_raw_extrinsic_metadata_swhid_authorities(request: Request, target: str):
    """
    .. http:get:: /api/1/raw-extrinsic-metadata/swhid/(target)/authorities/

        Returns a list of metadata authorities that provided metadata on
        the given target.
        They can then be used to get the raw `extrinsic metadata <https://docs.softwareheritage.org/devel/glossary.html#term-extrinsic-metadata>`__ collected on
        that object from each of the authorities.

        This endpoint should only be used directly to retrieve metadata from
        core SWHIDs (with type ``cnt``, ``dir``, ``rev``, ``rel``, and ``snp``).
        For "extended" SWHIDs such as origins,
        :http:get:`/api/1/raw-extrinsic-metadata/origin/(origin_url)/authorities/`
        should be used instead of building this URL directly.

        :param string target: The core SWHID of the object whose metadata-providing
          authorities should be returned

        {common_headers}

        :>jsonarr string type: Type of authority (deposit_client, forge, registry)
        :>jsonarr string url: Unique IRI identifying the authority
        :>jsonarr object metadata_list_url: URL to get the list of metadata objects
          on the given object from this authority

        :statuscode 200: no error

        **Example:**

        .. parsed-literal::

            :swh_web_api:`raw-extrinsic-metadata/swhid/swh:1:dir:a2faa28028657859c16ff506924212b33f0e1307/authorities/`
    """  # noqa

    try:
        parsed_target = swhids.ExtendedSWHID.from_string(target)
    except swhids.ValidationError as e:
        raise BadInputExc(f"Invalid target SWHID: {e}") from None

    authorities = config.storage().raw_extrinsic_metadata_get_authorities(
        target=parsed_target
    )
    results = [
        {
            **authority.to_dict(),
            "metadata_list_url": reverse(
                "api-1-raw-extrinsic-metadata-swhid",
                url_args={"target": target},
                query_params={"authority": f"{authority.type.value} {authority.url}"},
                request=request,
            ),
        }
        for authority in authorities
    ]

    return {
        "results": results,
        "headers": {},
    }


@api_route(
    "/raw-extrinsic-metadata/origin/(?P<origin_url>.*)/authorities/",
    "api-1-raw-extrinsic-metadata-origin-authorities",
)
@api_doc("/raw-extrinsic-metadata/origin/authorities/", category="Metadata")
@format_docstring()
def api_raw_extrinsic_metadata_origin_authorities(request: Request, origin_url: str):
    """
    .. http:get:: /api/1/raw-extrinsic-metadata/origin/(origin_url)/authorities/

        Similar to
        :http:get:`/api/1/raw-extrinsic-metadata/swhid/(target)/authorities/`
        but to get metadata on origins instead of objects

        :param string origin_url: The URL of the origin whose metadata-providing
          authorities should be returned

        {common_headers}

        :>jsonarr string type: Type of authority (deposit_client, forge, registry)
        :>jsonarr string url: Unique IRI identifying the authority
        :>jsonarr object metadata_list_url: URL to get the list of metadata objects
          on the given object from this authority

        :statuscode 200: no error

        **Example:**

        .. parsed-literal::

            :swh_web_api:`raw-extrinsic-metadata/origin/https://github.com/rdicosmo/parmap/authorities/`
    """  # noqa
    url = reverse(
        "api-1-raw-extrinsic-metadata-swhid-authorities",
        url_args={"target": Origin(url=origin_url).swhid()},
    )
    return redirect(url)
